from pydantic import root_validator

from gwproactor import ProactorSettings
from gwproactor.config import MQTTClient
from gwproactor_test.dummies.names import DUMMY_CHILD_ENV_PREFIX
from gwproactor_test.dummies.names import DUMMY_CHILD_NAME


class DummyChildSettings(ProactorSettings):
    parent_mqtt: MQTTClient = MQTTClient()
    seconds_per_report: int = 300
    async_power_reporting_threshold = 0.02

    class Config(ProactorSettings.Config):
        env_prefix = DUMMY_CHILD_ENV_PREFIX

    @root_validator(pre=True)
    def pre_root_validator(cls, values: dict) -> dict:
        return ProactorSettings.update_paths_name(values, DUMMY_CHILD_NAME)



from gwproactor.links.acks import DEFAULT_ACK_DELAY
from gwproactor.links.acks import AckManager
from gwproactor.links.acks import AckTimerCallback
from gwproactor.links.acks import AckWaitInfo
from gwproactor.links.asyncio_timer_manager import AsyncioTimerManager
from gwproactor.links.link_manager import LinkManager
from gwproactor.links.link_manager import LinkManagerTransition
from gwproactor.links.link_state import CommLinkAlreadyExists
from gwproactor.links.link_state import CommLinkMissing
from gwproactor.links.link_state import InvalidCommStateInput
from gwproactor.links.link_state import LinkState
from gwproactor.links.link_state import LinkStates
from gwproactor.links.link_state import RuntimeLinkStateError
from gwproactor.links.link_state import StateName
from gwproactor.links.link_state import Transition
from gwproactor.links.link_state import TransitionName
from gwproactor.links.message_times import LinkMessageTimes
from gwproactor.links.message_times import MessageTimes
from gwproactor.links.mqtt import QOS
from gwproactor.links.mqtt import MQTTClients
from gwproactor.links.mqtt import MQTTClientWrapper
from gwproactor.links.mqtt import Subscription
from gwproactor.links.reuploads import Reuploads
from gwproactor.links.timer_interface import TimerManagerInterface


__all__ = [
    "DEFAULT_ACK_DELAY",
    "AckManager",
    "AckTimerCallback",
    "AckWaitInfo",
    "AsyncioTimerManager",
    "CommLinkAlreadyExists",
    "CommLinkMissing",
    "InvalidCommStateInput",
    "LinkManager",
    "LinkManagerTransition",
    "LinkState",
    "LinkStates",
    "RuntimeLinkStateError",
    "Transition",
    "TransitionName",
    "StateName",
    "LinkMessageTimes",
    "MessageTimes",
    "MQTTClients",
    "MQTTClientWrapper",
    "QOS",
    "Reuploads",
    "Subscription",
    "TimerManagerInterface",
]



from typing import Optional

from gwproto import Decoders
from gwproto import Message
from gwproto import MQTTCodec
from gwproto import MQTTTopic
from gwproto import create_message_payload_discriminator

from gwproactor import ProactorSettings
from gwproactor.external_watchdog import SystemDWatchdogCommandBuilder
from gwproactor.links import QOS
from gwproactor.persister import TimedRollingFilePersister
from gwproactor.proactor_implementation import Proactor
from gwproactor_test.dummies.child.config import DummyChildSettings
from gwproactor_test.dummies.names import DUMMY_CHILD_NAME
from gwproactor_test.dummies.names import DUMMY_PARENT_NAME


ChildMessageDecoder = create_message_payload_discriminator(
    "ChildMessageDecoder",
    [
        "gwproto.messages",
        "gwproactor.message",
    ],
)


class ChildMQTTCodec(MQTTCodec):
    def __init__(self):
        super().__init__(
            Decoders.from_objects(message_payload_discriminator=ChildMessageDecoder)
        )

    def validate_source_alias(self, source_alias: str):
        if source_alias != DUMMY_PARENT_NAME:
            raise Exception(
                f"alias {source_alias} not my AtomicTNode ({DUMMY_PARENT_NAME})!"
            )


class DummyChild(Proactor):
    PARENT_MQTT = "gridworks"

    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyChildSettings] = None,
    ):
        super().__init__(
            name=name if name else DUMMY_CHILD_NAME,
            settings=DummyChildSettings() if settings is None else settings,
        )
        self._links.add_mqtt_link(
            DummyChild.PARENT_MQTT,
            settings.parent_mqtt,
            ChildMQTTCodec(),
            upstream=True,
            primary_peer=True,
        )
        for topic in [
            MQTTTopic.encode_subscription(Message.type_name(), DUMMY_PARENT_NAME),
            # Enable awaiting_setup edge case testing, which depends on receiving multiple, separate
            # MQTT topic subscription acks:
            MQTTTopic.encode_subscription(Message.type_name(), "1"),
            MQTTTopic.encode_subscription(Message.type_name(), "2"),
        ]:
            self._links.subscribe(self.PARENT_MQTT, topic, QOS.AtMostOnce)
        self._links.log_subscriptions("construction")

    @classmethod
    def make_event_persister(
        cls, settings: ProactorSettings
    ) -> TimedRollingFilePersister:
        return TimedRollingFilePersister(
            settings.paths.event_dir,
            pat_watchdog_args=SystemDWatchdogCommandBuilder.default_pat_args(),
        )

    @property
    def publication_name(self) -> str:
        return self.name

    @property
    def settings(self):
        return self._settings



"""Actor: A partial implementation of ActorInterface which supplies the trivial implementations.

SyncThreadActor: An actor which orchestrates starting, stopping and communicating with a passed in
SyncAsyncInteractionThread
"""

from abc import ABC
from typing import Any
from typing import Generic
from typing import Sequence
from typing import TypeVar

from gwproto import Message
from gwproto import ShNode
from result import Result

from gwproactor.proactor_interface import ActorInterface
from gwproactor.proactor_interface import Communicator
from gwproactor.proactor_interface import MonitoredName
from gwproactor.proactor_interface import ServicesInterface
from gwproactor.sync_thread import SyncAsyncInteractionThread


class Actor(ActorInterface, Communicator, ABC):

    _node: ShNode

    def __init__(self, name: str, services: ServicesInterface):
        self._node = services.hardware_layout.node(name)
        super().__init__(name, services)

    @property
    def alias(self):
        return self._name

    @property
    def node(self):
        return self._node

    def init(self) -> None:
        """Called after constructor so derived functions can be used in setup."""


SyncThreadT = TypeVar("SyncThreadT", bound=SyncAsyncInteractionThread)


class SyncThreadActor(Actor, Generic[SyncThreadT]):
    _sync_thread: SyncAsyncInteractionThread

    def __init__(
        self,
        name: str,
        services: ServicesInterface,
        sync_thread: SyncAsyncInteractionThread,
    ):
        super().__init__(name, services)
        self._sync_thread = sync_thread

    def process_message(self, message: Message) -> Result[bool, BaseException]:
        raise ValueError(
            f"Error. {self.__class__.__name__} does not process any messages. Received {message.Header}"
        )

    def send_driver_message(self, message: Any) -> None:
        self._sync_thread.put_to_sync_queue(message)

    def start(self):
        self._sync_thread.set_async_loop_and_start(
            self.services.event_loop, self.services.async_receive_queue
        )

    def stop(self):
        self._sync_thread.request_stop()

    async def join(self):
        await self._sync_thread.async_join()

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        monitored_names = []
        if self._sync_thread.pat_timeout is not None:
            monitored_names.append(
                MonitoredName(self.name, self._sync_thread.pat_timeout)
            )
        return monitored_names



"""MQTT infrastructure providing support for multiple MTQTT clients

TODO: Replace synchronous use of Paho MQTT Client with asyncio usage, per Paho documentation or external library

Main current limitation: each interaction between asyncio code and the mqtt clients must either have thread locking
(as is provided inside paho for certain functions such as publish()) or an explicit message based API.

"""

import asyncio
import enum
import logging
import ssl
import threading
import uuid
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union
from typing import cast

from paho.mqtt.client import MQTT_ERR_SUCCESS
from paho.mqtt.client import Client as PahoMQTTClient
from paho.mqtt.client import MQTTMessageInfo

from gwproactor import config
from gwproactor.message import MQTTConnectFailMessage
from gwproactor.message import MQTTConnectMessage
from gwproactor.message import MQTTDisconnectMessage
from gwproactor.message import MQTTProblemsMessage
from gwproactor.message import MQTTReceiptMessage
from gwproactor.message import MQTTSubackMessage
from gwproactor.message import MQTTSubackPayload
from gwproactor.problems import Problems
from gwproactor.sync_thread import AsyncQueueWriter
from gwproactor.sync_thread import responsive_sleep


class QOS(enum.IntEnum):
    AtMostOnce = 0
    AtLeastOnce = 1
    ExactlyOnce = 2


class Subscription(NamedTuple):
    Topic: str
    Qos: QOS


class MQTTClientWrapper:
    _name: str
    _client_config: config.MQTTClient
    _client: PahoMQTTClient
    _stop_requested: bool
    _receive_queue: AsyncQueueWriter
    _subscriptions: Dict[str, int]
    _pending_subscriptions: Set[str]
    _pending_subacks: Dict[int, List[str]]

    def __init__(
        self,
        name: str,
        client_config: config.MQTTClient,
        receive_queue: AsyncQueueWriter,
    ):
        self.name = name
        self._client_config = client_config
        self._receive_queue = receive_queue
        self._client = PahoMQTTClient("-".join(str(uuid.uuid4()).split("-")[:-1]))
        self._client.username_pw_set(
            username=self._client_config.username,
            password=self._client_config.password.get_secret_value(),
        )
        tls_config = self._client_config.tls
        if tls_config.use_tls:
            self._client.tls_set(
                ca_certs=tls_config.paths.ca_cert_path,
                certfile=tls_config.paths.cert_path,
                keyfile=tls_config.paths.private_key_path,
                cert_reqs=tls_config.cert_reqs,
                tls_version=ssl.PROTOCOL_TLS_CLIENT,
                ciphers=tls_config.ciphers,
                keyfile_password=tls_config.keyfile_password.get_secret_value(),
            )

        self._client.on_message = self.on_message
        self._client.on_connect = self.on_connect
        self._client.on_connect_fail = self.on_connect_fail
        self._client.on_disconnect = self.on_disconnect
        self._client.on_subscribe = self.on_subscribe
        self._subscriptions = dict()
        self._pending_subscriptions = set()
        self._pending_subacks = dict()
        self._thread = threading.Thread(
            target=self._client_thread, name=f"MQTT-client-thread-{self.name}"
        )
        self._stop_requested = False

    def _client_thread(self):
        MAX_BACK_OFF = 1024
        backoff = 1
        while not self._stop_requested:
            try:
                self._client.connect(
                    self._client_config.host, port=self._client_config.effective_port()
                )
                self._client.loop_forever(retry_first_connection=True)
            except BaseException as e:
                self._receive_queue.put(
                    MQTTProblemsMessage(
                        client_name=self.name, problems=Problems(errors=[e])
                    )
                )
            finally:
                # noinspection PyBroadException
                try:
                    self._client.disconnect()
                except:
                    pass
            if not self._stop_requested:
                if backoff >= MAX_BACK_OFF:
                    backoff = 1
                else:
                    backoff = min(backoff * 2, MAX_BACK_OFF)
                responsive_sleep(
                    self,
                    backoff,
                    running_field_name="_stop_requested",
                    running_field=False,
                )

    def start(self):
        self._thread.start()

    def stop(self):
        self._stop_requested = True
        try:
            self._client.disconnect()
            self._thread.join()
        except:  # noqa
            pass

    def publish(self, topic: str, payload: bytes, qos: int) -> MQTTMessageInfo:
        return self._client.publish(topic, payload, qos)

    def subscribe(self, topic: str, qos: int) -> Tuple[int, Optional[int]]:
        self._subscriptions[topic] = qos
        self._pending_subscriptions.add(topic)
        subscribe_result = self._client.subscribe(topic, qos)
        if subscribe_result[0] == MQTT_ERR_SUCCESS:
            self._pending_subacks[subscribe_result[1]] = [topic]
        return subscribe_result

    def subscribe_all(self) -> Tuple[int, Optional[int]]:
        if self._subscriptions:
            topics = list(self._subscriptions.keys())
            for topic in topics:
                self._pending_subscriptions.add(topic)
            subscribe_result = self._client.subscribe(
                list(self._subscriptions.items()), 0
            )
            if subscribe_result[0] == MQTT_ERR_SUCCESS:
                self._pending_subacks[subscribe_result[1]] = topics
        else:
            subscribe_result = MQTT_ERR_SUCCESS, None
        return subscribe_result

    def unsubscribe(self, topic: str) -> Tuple[int, Optional[int]]:
        self._subscriptions.pop(topic, None)
        return self._client.unsubscribe(topic)

    def connected(self) -> bool:
        return self._client.is_connected()

    def num_subscriptions(self) -> int:
        return len(self._subscriptions)

    def num_pending_subscriptions(self) -> int:
        return len(self._pending_subscriptions)

    def subscribed(self) -> bool:
        return self.connected() and (
            not self.num_subscriptions() or not self.num_pending_subscriptions()
        )

    def subscription_items(self) -> list[Tuple[str, int]]:
        return list(cast(list[Tuple[str, int]], self._subscriptions.items()))

    def on_message(self, _, userdata, message):
        self._receive_queue.put(
            MQTTReceiptMessage(
                client_name=self.name,
                userdata=userdata,
                message=message,
            )
        )

    def handle_suback(self, suback: MQTTSubackPayload) -> int:
        topics = self._pending_subacks.pop(suback.mid, [])
        if topics:
            for topic in topics:
                self._pending_subscriptions.remove(topic)
        return len(self._pending_subscriptions)

    def on_subscribe(self, _, userdata, mid, granted_qos):
        self._receive_queue.put(
            MQTTSubackMessage(
                client_name=self.name,
                userdata=userdata,
                mid=mid,
                granted_qos=granted_qos,
            )
        )

    def on_connect(self, _, userdata, flags, rc):
        self._receive_queue.put(
            MQTTConnectMessage(
                client_name=self.name,
                userdata=userdata,
                flags=flags,
                rc=rc,
            )
        )

    def on_connect_fail(self, _, userdata):
        self._receive_queue.put(
            MQTTConnectFailMessage(
                client_name=self.name,
                userdata=userdata,
            )
        )

    def on_disconnect(self, _, userdata, rc):
        self._pending_subscriptions = set(self._subscriptions.keys())
        self._receive_queue.put(
            MQTTDisconnectMessage(
                client_name=self.name,
                userdata=userdata,
                rc=rc,
            )
        )

    def enable_logger(
        self, logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None
    ):
        self._client.enable_logger(logger)

    def disable_logger(self):
        self._client.disable_logger()


class MQTTClients:
    clients: Dict[str, MQTTClientWrapper]
    _send_queue: AsyncQueueWriter
    upstream_client: str = ""
    primary_peer_client: str = ""

    def __init__(self):
        self._send_queue = AsyncQueueWriter()
        self.clients = dict()

    def add_client(
        self,
        name: str,
        client_config: config.MQTTClient,
        upstream: bool = False,
        primary_peer: bool = False,
    ):
        if name in self.clients:
            raise ValueError(f"ERROR. MQTT client named {name} already exists")
        if upstream:
            if self.upstream_client:
                raise ValueError(
                    f"ERROR. upstream client already set as {self.upstream_client}. Client {name} may not be set as upstream."
                )
            self.upstream_client = name
        if primary_peer:
            if self.primary_peer_client:
                raise ValueError(
                    f"ERROR. primary peer client already set as {self.primary_peer_client}. Client {name} may not be set as primary peer."
                )
            self.primary_peer_client = name
        self.clients[name] = MQTTClientWrapper(name, client_config, self._send_queue)

    def publish(
        self, client: str, topic: str, payload: bytes, qos: int
    ) -> MQTTMessageInfo:
        return self.clients[client].publish(topic, payload, qos)

    def subscribe(self, client: str, topic: str, qos: int) -> Tuple[int, Optional[int]]:
        return self.clients[client].subscribe(topic, qos)

    def subscribe_all(self, client: str) -> Tuple[int, Optional[int]]:
        return self.clients[client].subscribe_all()

    def unsubscribe(self, client: str, topic: str) -> Tuple[int, Optional[int]]:
        return self.clients[client].unsubscribe(topic)

    def handle_suback(self, suback: MQTTSubackPayload) -> int:
        return self.clients[suback.client_name].handle_suback(suback)

    def stop(self):
        for client in self.clients.values():
            client.stop()

    def start(self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue):
        self._send_queue.set_async_loop(loop, async_queue)
        for client in self.clients.values():
            client.start()

    def connected(self, client: str) -> bool:
        return self.clients[client].connected()

    def num_subscriptions(self, client: str) -> int:
        return self.clients[client].num_subscriptions()

    def num_pending_subscriptions(self, client: str) -> int:
        return self.clients[client].num_pending_subscriptions()

    def subscribed(self, client: str) -> bool:
        return self.clients[client].subscribed()

    def enable_loggers(
        self, logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None
    ):
        for client_name in self.clients:
            self.clients[client_name].enable_logger(logger)

    def disable_loggers(self):
        for client_name in self.clients:
            self.clients[client_name].disable_logger()

    def client_wrapper(self, client: str) -> MQTTClientWrapper:
        return self.clients[client]

    def upstream(self) -> MQTTClientWrapper:
        return self.clients[self.upstream_client]

    def primary_peer(self) -> MQTTClientWrapper:
        return self.clients[self.primary_peer_client]



"""A 'reupload' is a process of uploading events that have not yet been acked when communication with the upstream
peer is re-activated. The re-upload proceeds slowly by receiving acks so as not to overwhelm processing queues. Events
that occur after communication has been reactivated and/or during a re-upload are not part of the reupload, but are sent
as they occur.

This module provides a class, Reuploads, for tracking which events are part of the reupload.

This module only manages events ids; it does not change (add or remove) event storage.
"""

from typing import Optional

from gwproactor.logger import ProactorLogger
from gwproactor.stats import LinkStats


class _ReuploadDiffLogger:  # pragma: no cover
    """Helper class for logging results of an ack without to much logging code bulk in the ack processing routine"""

    reuploads: Optional["Reuploads"] = None
    event_id: str = ""
    verbose: bool = False
    begin_reuploading: bool = False
    begin_num_unacked: int = -1
    begin_num_pending: int = -1
    begin_verbose_str: str = ""

    def init(
        self,
        reuploads: Optional["Reuploads"] = None,
        verbose: bool = False,
    ):
        self.reuploads = reuploads
        self.verbose = verbose
        if reuploads is not None:
            self.begin_reuploading = reuploads.reuploading()
            self.begin_num_unacked = reuploads.num_reuploaded_unacked
            self.begin_num_pending = reuploads.num_reupload_pending
            if verbose:
                self.begin_verbose_str = reuploads.get_str(num_events=5)

    def diff_str(self, path_dbg: int) -> str:
        s = ""
        if self.reuploads is not None:
            if self.verbose:
                s += f"Begin reuploads:\n{self.begin_verbose_str}\n"
                s += f"End reuploads:\n{self.reuploads.get_str(num_events=5)}\n"
            s += (
                f"path:0x{path_dbg:08X}  "
                f"reuploading: {int(self.begin_reuploading)} -> {int(self.reuploads.reuploading())}  "
                f"unacked: {self.begin_num_unacked} -> {self.reuploads.num_reuploaded_unacked}  "
                f"pending: {self.begin_num_pending} -> {self.reuploads.num_reupload_pending}  "
            )
        return s

    def ack_str(self, path_dbg: int) -> str:
        return f"--process_ack_for_reupload  {self.diff_str(path_dbg)}"

    def log_ack(self, path_dbg: int) -> None:
        if self.reuploads is not None and self.reuploads._logger.path_enabled:  # noqa
            self.reuploads._logger.path(self.ack_str(path_dbg))  # noqa


class Reuploads:
    """Track event uids that are part of a re-upload, both those that have not yet been sent (pending) and those that are
    "in-flight", (unacked). Upon ack, update records and return next message to be sent.
    """

    NUM_INITIAL_EVENTS: int = 5
    """Default number of events to send when re-upload starts."""

    _reupload_pending: dict[str, None]
    """'Ordered set' of *unsent* events that are part of this reupload.
    A dict is used to provide insertion order and fast lookup."""

    _reuploaded_unacked: dict[str, None]
    """'Ordered set' of *sent but as-yet unacked*
    A dict is used to provide insertion order and fast lookup."""

    _num_initial_events: int
    """Number of events to send when re-upload starts."""

    stats: Optional[LinkStats] = None
    """Object into which we can record reupload start and complete. Set during
    LinkManager.start() since upstream client does not exist durin LinkManager
    construction. """

    _logger: ProactorLogger

    def __init__(
        self,
        logger: ProactorLogger,
        num_initial_events: int = NUM_INITIAL_EVENTS,
    ):
        self._reupload_pending = dict()
        self._reuploaded_unacked = dict()
        self._num_initial_events = num_initial_events
        self._logger = logger

    def start_reupload(self, pending_events: list[str]) -> list[str]:
        """Track all pending_events for reupload. Of the pending_events,
        record the first _num_initial_events as "unacked" and the rest
        as "pending". Return the "unacked" group so they can be sent.
        """
        reupload_now = pending_events[: self._num_initial_events]
        self._reuploaded_unacked = dict.fromkeys(reupload_now)
        self._reupload_pending = dict.fromkeys(
            pending_events[self._num_initial_events :]
        )
        if self.reuploading():
            self.stats.start_reupload()
        self._log_start_reupload(len(pending_events), len(reupload_now))
        return reupload_now

    def clear_unacked_event(self, ack_id: str) -> None:
        self._reuploaded_unacked.pop(ack_id)

    def process_ack_for_reupload(self, ack_id: str) -> list[str]:
        """If ack_id is in our "unacked" store, remove it from the unacked store. If any events remain in our "pending"
        store, move the first pending event from the pending store to the unacked store and return it for sending
        next."""

        path_dbg = 0
        was_reuploading = self.reuploading()
        ack_logger = _ReuploadDiffLogger()
        if self._logger.path_enabled:
            ack_logger.init(self, verbose=True)
        reupload_now = []
        if ack_id in self._reuploaded_unacked:
            path_dbg |= 0x00000001
            self._reuploaded_unacked.pop(ack_id)
            if self._reupload_pending:
                path_dbg |= 0x00000002
                reupload_next = next(iter(self._reupload_pending))
                self._reupload_pending.pop(reupload_next)
                self._reuploaded_unacked[reupload_next] = None
                reupload_now = [reupload_next]
        # This case is likely in testing (which explicitly generates
        # the awaiting_setup state), but unlikely in the real works, since
        # unless we have many subscriptions we will get one suback for all of
        # them, meaning we can't enter the awaiting_setup state.
        # In awaiting_setup, this case is likely since we generate an event for
        # the suback before we start the reupload - we generate the event, send it,
        # then add it to the reupload. Because old events are sent first in the reupload,
        # the suback event is likely to be in pending, not in unacked.
        # In theory this could also happen if an ack for an event sent prior to
        # communication loss was somehow preserved in a queue and delivered after comm
        # restore.
        elif ack_id in self._reupload_pending:
            path_dbg |= 0x00000004
            self._reupload_pending.pop(ack_id)
        if was_reuploading and not self.reuploading():
            path_dbg |= 0x00000008
            self.stats.complete_reupload()
        if self._logger.path_enabled:
            ack_logger.log_ack(path_dbg)
        return reupload_now

    @property
    def num_reupload_pending(self) -> int:
        return len(self._reupload_pending)

    @property
    def num_reuploaded_unacked(self) -> int:
        return len(self._reuploaded_unacked)

    def reuploading(self) -> bool:
        return bool(len(self._reuploaded_unacked) + len(self._reupload_pending))

    def clear(self) -> None:
        self._reupload_pending.clear()
        self._reuploaded_unacked.clear()

    def get_str(self, verbose: bool = True, num_events: int = 5) -> str:
        s = f"Reuploads  reuploading:{int(self.reuploading())}  unacked/sent:{len(self._reuploaded_unacked)}  pending/unsent:{len(self._reupload_pending)}"
        if verbose:
            s += f"  num initial:{self._num_initial_events}\n"
            s += f"  unacked:{len(self._reuploaded_unacked)}\n"
            for message_id in self._reuploaded_unacked:
                s += f"    {message_id[:8]}...\n"
            s += f"  pending:{len(self._reupload_pending)}\n"
            for i, message_id in enumerate(self._reupload_pending):
                s += f"    {message_id[:8]}...\n"
                if i == num_events - 1:
                    break
        return s.rstrip()

    def __str__(self):
        return self.get_str(verbose=False)

    def _log_start_reupload(self, num_pending_events, num_reupload_now):
        if self._logger.general_enabled:
            if self.reuploading():
                state_str = f"{self.num_reupload_pending} reupload events pending."
            else:
                state_str = "reupload complete."
            self._logger.info(
                f"start_reupload: sent {num_reupload_now} events. "
                f"{state_str} "
                f"Total events in reupload: {num_pending_events}."
            )
        if self._logger.path_enabled:
            self._logger.path(self.get_str(num_events=5))



"""Scada implementation"""

from typing import Optional
from typing import cast

from gwproto import Decoders
from gwproto import MQTTCodec
from gwproto import MQTTTopic
from gwproto import create_message_payload_discriminator

from gwproactor.links import QOS
from gwproactor.message import Message
from gwproactor.persister import SimpleDirectoryWriter
from gwproactor.proactor_implementation import Proactor
from gwproactor_test.dummies.names import DUMMY_CHILD_NAME
from gwproactor_test.dummies.names import DUMMY_PARENT_NAME
from gwproactor_test.dummies.parent.config import DummyParentSettings


ParentMessageDecoder = create_message_payload_discriminator(
    model_name="ParentMessageDecoder",
    module_names=["gwproto.messages", "gwproactor.message"],
)


class ParentMQTTCodec(MQTTCodec):
    def __init__(self):
        super().__init__(
            Decoders.from_objects(message_payload_discriminator=ParentMessageDecoder)
        )

    def validate_source_alias(self, source_alias: str):
        if source_alias != DUMMY_CHILD_NAME:
            raise Exception(f"alias {source_alias} not my Scada!")


class DummyParent(Proactor):
    CHILD_MQTT = "child"

    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyParentSettings] = None,
    ):
        super().__init__(
            name=name if name else DUMMY_PARENT_NAME,
            settings=DummyParentSettings() if settings is None else settings,
        )
        self._links.add_mqtt_link(
            self.CHILD_MQTT,
            self.settings.child_mqtt,
            ParentMQTTCodec(),
            primary_peer=True,
        )
        self._links.subscribe(
            self.CHILD_MQTT,
            MQTTTopic.encode_subscription(Message.type_name(), DUMMY_CHILD_NAME),
            QOS.AtMostOnce,
        )

    @classmethod
    def make_event_persister(
        cls, settings: DummyParentSettings
    ) -> SimpleDirectoryWriter:
        return SimpleDirectoryWriter(settings.paths.event_dir)

    @property
    def publication_name(self) -> str:
        return self.name

    @property
    def settings(self) -> DummyParentSettings:
        return cast(DummyParentSettings, self._settings)



import asyncio
import json
import logging
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

from gwproto import Message
from gwproto import MQTTCodec
from gwproto.messages import Ack
from gwproto.messages import CommEvent
from gwproto.messages import EventT
from gwproto.messages import MQTTConnectEvent
from gwproto.messages import MQTTDisconnectEvent
from gwproto.messages import MQTTFullySubscribedEvent
from gwproto.messages import PeerActiveEvent
from gwproto.messages import PingMessage
from gwproto.messages import ProblemEvent
from gwproto.messages import ResponseTimeoutEvent
from gwproto.messages import StartupEvent
from paho.mqtt.client import MQTTMessageInfo
from result import Err
from result import Ok
from result import Result

from gwproactor.config import MQTTClient
from gwproactor.config import ProactorSettings
from gwproactor.links import AckWaitInfo
from gwproactor.links.acks import AckManager
from gwproactor.links.acks import AckTimerCallback
from gwproactor.links.link_state import InvalidCommStateInput
from gwproactor.links.link_state import LinkState
from gwproactor.links.link_state import LinkStates
from gwproactor.links.link_state import StateName
from gwproactor.links.link_state import Transition
from gwproactor.links.message_times import LinkMessageTimes
from gwproactor.links.message_times import MessageTimes
from gwproactor.links.mqtt import QOS
from gwproactor.links.mqtt import MQTTClients
from gwproactor.links.mqtt import MQTTClientWrapper
from gwproactor.links.reuploads import Reuploads
from gwproactor.links.timer_interface import TimerManagerInterface
from gwproactor.logger import ProactorLogger
from gwproactor.message import MQTTConnectFailPayload
from gwproactor.message import MQTTConnectPayload
from gwproactor.message import MQTTDisconnectPayload
from gwproactor.message import MQTTReceiptPayload
from gwproactor.message import MQTTSubackPayload
from gwproactor.persister import ByteDecodingError
from gwproactor.persister import DecodingError
from gwproactor.persister import FileEmptyWarning
from gwproactor.persister import JSONDecodingError
from gwproactor.persister import PersisterInterface
from gwproactor.persister import UIDMissingWarning
from gwproactor.problems import Problems
from gwproactor.stats import ProactorStats


@dataclass
class LinkManagerTransition(Transition):
    canceled_acks: list[AckWaitInfo] = field(default_factory=list)


class LinkManager:
    PERSISTER_ENCODING = "utf-8"
    publication_name: str
    _settings: ProactorSettings
    _logger: ProactorLogger
    _stats: ProactorStats
    _event_persister: PersisterInterface
    _reuploads: Reuploads
    _mqtt_clients: MQTTClients
    _mqtt_codecs: dict[str, MQTTCodec]
    _states: LinkStates
    _message_times: MessageTimes
    _acks: AckManager

    def __init__(
        self,
        publication_name: str,
        settings: ProactorSettings,
        logger: ProactorLogger,
        stats: ProactorStats,
        event_persister: PersisterInterface,
        timer_manager: TimerManagerInterface,
        ack_timeout_callback: AckTimerCallback,
    ):
        self.publication_name = publication_name
        self._settings = settings
        self._logger = logger
        self._stats = stats
        self._event_persister = event_persister
        self._reuploads = Reuploads(
            self._logger,
            self._settings.num_initial_event_reuploads,
        )
        self._mqtt_clients = MQTTClients()
        self._mqtt_codecs = dict()
        self._states = LinkStates()
        self._message_times = MessageTimes()
        self._acks = AckManager(
            timer_manager, ack_timeout_callback, delay=settings.ack_timeout_seconds
        )

    @property
    def upstream_client(self) -> str:
        return self._mqtt_clients.upstream_client

    @property
    def primary_peer_client(self) -> str:
        return self._mqtt_clients.primary_peer_client

    @property
    def num_pending(self) -> int:
        return self._event_persister.num_pending

    @property
    def num_reupload_pending(self) -> int:
        return self._reuploads.num_reupload_pending

    @property
    def num_reuploaded_unacked(self) -> int:
        return self._reuploads.num_reuploaded_unacked

    def reuploading(self) -> bool:
        return self._reuploads.reuploading()

    def num_acks(self, link_name: str) -> int:
        return self._acks.num_acks(link_name)

    def subscribed(self, link_name: str) -> bool:
        return self._mqtt_clients.subscribed(link_name)

    def mqtt_clients(self) -> MQTTClients:
        return self._mqtt_clients

    def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper:
        return self._mqtt_clients.client_wrapper(client_name)

    def enable_mqtt_loggers(
        self, logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None
    ):
        return self._mqtt_clients.enable_loggers(logger)

    def disable_mqtt_loggers(self):
        return self._mqtt_clients.disable_loggers()

    def decoder(self, link_name: str) -> Optional[MQTTCodec]:
        return self._mqtt_codecs.get(link_name, None)

    def decode(self, link_name: str, topic: str, payload: bytes) -> Message[Any]:
        return self._mqtt_codecs[link_name].decode(topic, payload)

    def link(self, name) -> Optional[LinkState]:
        return self._states.link(name)

    def link_state(self, name) -> Optional[StateName]:
        return self._states.link_state(name)

    def link_names(self) -> list[str]:
        return self._states.link_names()

    def __contains__(self, name: str) -> bool:
        return name in self._states

    def __getitem__(self, name: str) -> LinkState:
        return self._states[name]

    def get_message_times(self, link_name: str) -> LinkMessageTimes:
        return self._message_times.get_copy(link_name)

    def stopped(self, name: str) -> bool:
        return self._states.stopped(name)

    def add_mqtt_link(
        self,
        name: str,
        mqtt_config: MQTTClient,
        codec: Optional[MQTTCodec] = None,
        upstream: bool = False,
        primary_peer: bool = False,
    ):
        self._mqtt_clients.add_client(
            name,
            mqtt_config,
            upstream=upstream,
            primary_peer=primary_peer,
        )
        if codec is not None:
            self._mqtt_codecs[name] = codec
        self._states.add(name)
        self._message_times.add_link(name)
        self._stats.add_link(name)

    def subscribe(self, client: str, topic: str, qos: int) -> Tuple[int, Optional[int]]:
        return self._mqtt_clients.subscribe(client, topic, qos)

    def log_subscriptions(self, tag=""):
        if self._logger.lifecycle_enabled:
            s = f"Subscriptions: [{tag}]]\n"
            for client in self._mqtt_clients.clients:
                s += f"\t{client}\n"
                for subscription in self._mqtt_clients.client_wrapper(
                    client
                ).subscription_items():
                    s += f"\t\t[{subscription}]\n"
            self._logger.lifecycle(s)

    def get_reuploads_str(self, verbose: bool = True, num_events: int = 5) -> str:
        return self._reuploads.get_str(verbose=verbose, num_events=num_events)

    def publish_message(
        self, client, message: Message, qos: int = 0, context: Any = None
    ) -> MQTTMessageInfo:
        topic = message.mqtt_topic()
        payload = self._mqtt_codecs[client].encode(message)
        self._logger.message_summary(
            "OUT mqtt    ",
            message.Header.Src,
            topic,
            message.Payload,
            message_id=message.Header.MessageId,
        )
        if message.Header.AckRequired:
            self._acks.start_ack_timer(
                client, message.Header.MessageId, context=context
            )
        self._message_times.update_send(client)
        return self._mqtt_clients.publish(client, topic, payload, qos)

    def publish_upstream(
        self, payload, qos: QOS = QOS.AtMostOnce, **message_args: Any
    ) -> MQTTMessageInfo:
        message = Message(Src=self.publication_name, Payload=payload, **message_args)
        return self.publish_message(
            self._mqtt_clients.upstream_client, message, qos=qos
        )

    def generate_event(self, event: EventT) -> Result[bool, BaseException]:
        if isinstance(event, CommEvent):
            self._stats.link(event.PeerName).comm_event_counts[event.TypeName] += 1
        if isinstance(event, ProblemEvent) and self._logger.path_enabled:
            self._logger.info(event)
        if not event.Src:
            event.Src = self.publication_name
        if (
            self._mqtt_clients.upstream_client
            and self._states[self._mqtt_clients.upstream_client].active_for_send()
        ):
            self.publish_upstream(event, AckRequired=True)
        return self._event_persister.persist(
            event.MessageId,
            event.json(sort_keys=True, indent=2).encode(self.PERSISTER_ENCODING),
        )

    def _start_reupload(self) -> None:
        if not self._reuploads.reuploading():
            self._continue_reupload(
                self._reuploads.start_reupload(self._event_persister.pending())
            )

    def _continue_reupload(self, event_ids: list[str]) -> None:
        self._logger.path("++_continue_reupload  %d", len(event_ids))
        path_dbg = 0
        tried_count_dbg = 0
        sent_count_dbg = 0
        continuation_count_dbg = -1

        if event_ids:
            path_dbg |= 0x00000001
            sent_one = False
            # Try to send all requested events. At least send must succeed to
            # continue the reupload, so if all sends fail, get more until
            # one is sent or there are no more reuploads.
            while not sent_one and self._reuploads.reuploading() and event_ids:
                continuation_path_dbg = 0x00000002
                continuation_count_dbg += 1
                next_event_ids = []
                for event_id in event_ids:
                    event_path_dbg = 0x00000004
                    tried_count_dbg += 1
                    problems = Problems()
                    ret = self._reupload_event(event_id)
                    if ret.is_ok():
                        event_path_dbg |= 0x00000008
                        if ret.value:
                            path_dbg |= 0x00000010
                            sent_count_dbg += 1
                            sent_one = True
                        else:
                            event_path_dbg |= 0x00000020
                            problems.add_error(DecodingError(uid=event_id))
                    else:
                        event_path_dbg |= 0x00000040
                        problems.add_problems(ret.err())
                    if problems:
                        event_path_dbg |= 0x00000080
                        # There was some error decoding this event.
                        # We generate a new event with information
                        # about decoding failure and delete this event.
                        self.generate_event(
                            problems.problem_event(
                                f"Event decoding error - uid:{event_id}"
                            )
                        )
                        self._event_persister.clear(event_id)
                        if sent_one:
                            event_path_dbg |= 0x00000100
                            self._reuploads.clear_unacked_event(event_id)
                        else:
                            event_path_dbg |= 0x00000200
                            next_event_ids.extend(
                                self._reuploads.process_ack_for_reupload(event_id)
                            )
                    self._logger.path("  1 event path:0x%08X", event_path_dbg)
                    continuation_path_dbg |= event_path_dbg
                self._logger.path("  1 continuation path:0x%08X", continuation_path_dbg)
                event_ids = next_event_ids
                path_dbg |= continuation_path_dbg
        self._logger.path(
            "--_continue_reupload  path:0x%08X  sent:%d  tried:%d  continuations:%d",
            path_dbg,
            sent_count_dbg,
            tried_count_dbg,
            continuation_count_dbg,
        )

    def _reupload_event(self, event_id) -> Result[bool, Problems]:
        """Load event for event_id from storage, decoded to JSON and send it.

        Return either Ok(True) or Err(Problems(list of decoding errors)).

        Send errors handled either by exception, which will propagate up, or
        by ack timeout.
        """
        self._logger.path("++_reupload_event  %s", event_id)
        path_dbg = 0
        problems = Problems()
        match self._event_persister.retrieve(event_id):
            case Ok(event_bytes):
                path_dbg |= 0x00000001
                if event_bytes is None:
                    path_dbg |= 0x00000002
                    problems.add_error(
                        UIDMissingWarning("reupload_events", uid=event_id)
                    )
                elif len(event_bytes) == 0:
                    path_dbg |= 0x00000004
                    problems.add_error(
                        FileEmptyWarning("reupload_events", uid=event_id)
                    )
                else:
                    path_dbg |= 0x00000008
                    try:
                        event_str = event_bytes.decode(encoding=self.PERSISTER_ENCODING)
                    except BaseException as e:
                        path_dbg |= 0x00000010
                        problems.add_error(e).add_error(
                            ByteDecodingError("reupload_events", uid=event_id)
                        )
                    else:
                        path_dbg |= 0x00000020
                        try:
                            event = json.loads(event_str)
                        except BaseException as e:
                            path_dbg |= 0x00000040
                            problems.add_error(e).add_error(
                                JSONDecodingError(
                                    f"reupload_events - raw json:\n<\n{event_str}\n>",
                                    uid=event_id,
                                )
                            )
                        else:
                            path_dbg |= 0x00000080
                            self.publish_upstream(event, AckRequired=True)
                            self._logger.path(
                                "--_reupload_event:1  path:0x%08X", path_dbg
                            )
                            return Ok(True)
            case Err(error):
                path_dbg |= 0x00000100
                problems.add_problems(error)
        self._logger.path("--_reupload_event:0  path:0x%08X", path_dbg)
        return Err(problems)

    def start(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue
    ) -> None:
        if self.upstream_client:
            self._reuploads.stats = self._stats.link(self.upstream_client)
        self._mqtt_clients.start(loop, async_queue)
        self.generate_event(StartupEvent())
        self._states.start_all()

    def stop(self) -> Result[bool, Problems]:
        problems: Optional[Problems] = None
        for link_name in self._states.link_names():
            ret = self._states.stop(link_name)
            if ret.is_err():
                if problems is None:
                    problems = Problems(errors=[ret.err()])
                else:
                    problems.errors.append(ret.err())
        self._mqtt_clients.stop()
        if problems is None:
            return Ok(True)
        return Err(problems)

    def process_mqtt_connected(
        self, message: Message[MQTTConnectPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        state_result = self._states.process_mqtt_connected(message)
        if state_result.is_ok():
            self._logger.comm_event(str(state_result.value))
        self.generate_event(MQTTConnectEvent(PeerName=message.Payload.client_name))
        self._mqtt_clients.subscribe_all(message.Payload.client_name)
        return state_result

    def process_mqtt_disconnected(
        self, message: Message[MQTTDisconnectPayload]
    ) -> Result[LinkManagerTransition, InvalidCommStateInput]:
        state_result = self._states.process_mqtt_disconnected(message)
        if state_result.is_ok():
            result = Ok(LinkManagerTransition(**(asdict(state_result.value))))
            self.generate_event(
                MQTTDisconnectEvent(PeerName=message.Payload.client_name)
            )
            self._logger.comm_event(str(result.value))
            if result.value.recv_deactivated() or result.value.send_deactivated():
                result.value.canceled_acks = self._acks.cancel_ack_timers(
                    message.Payload.client_name
                )
                self._reuploads.clear()
        else:
            result = state_result
        return result

    def process_mqtt_connect_fail(
        self, message: Message[MQTTConnectFailPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        return self._states.process_mqtt_connect_fail(message)

    def process_mqtt_message(
        self, message: Message[MQTTReceiptPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        result = self._states.process_mqtt_message(message)
        if result.is_ok():
            self.update_recv_time(message.Payload.client_name)
        if result.value:
            self._logger.comm_event(str(result.value))
        if result.value.recv_activated():
            self._recv_activated(result.value)
        return result

    def process_ack_timeout(
        self, wait_info: AckWaitInfo
    ) -> Result[LinkManagerTransition, BaseException]:
        self._logger.path("++LinkManager.process_ack_timeout %s", wait_info.message_id)
        path_dbg = 0
        self._stats.link(wait_info.link_name).timeouts += 1
        state_result = self._states.process_ack_timeout(wait_info.link_name)
        if state_result.is_ok():
            result = Ok(
                LinkManagerTransition(
                    canceled_acks=[wait_info], **(asdict(state_result.value))
                )
            )
            path_dbg |= 0x00000001
            if result.value.deactivated():
                path_dbg |= 0x00000002
                self._reuploads.clear()
                self.generate_event(
                    ResponseTimeoutEvent(PeerName=result.value.link_name)
                )
                self._logger.comm_event(str(result.value))
                result.value.canceled_acks.extend(
                    self._acks.cancel_ack_timers(wait_info.link_name)
                )
        else:
            result = Err(state_result.err())
        self._logger.path("--LinkManager.process_ack_timeout path:0x%08X", path_dbg)
        return result

    def process_ack(self, link_name: str, message_id: str):
        self._logger.path("++LinkManager.process_ack  %s", message_id)
        path_dbg = 0
        wait_info = self._acks.cancel_ack_timer(link_name, message_id)
        if wait_info is not None and message_id in self._event_persister:
            path_dbg |= 0x00000001
            self._event_persister.clear(message_id)
            if self._reuploads.reuploading() and link_name == self.upstream_client:
                path_dbg |= 0x00000002
                self._continue_reupload(
                    self._reuploads.process_ack_for_reupload(message_id)
                )
                if not self._reuploads.reuploading():
                    path_dbg |= 0x00000004
                    self._logger.info("reupload complete.")
        self._logger.path("--LinkManager.process_ack path:0x%08X", path_dbg)

    def send_ack(self, link_name: str, message: Message[Any]) -> None:
        if message.Header.MessageId:
            self.publish_message(
                link_name,
                Message(
                    Src=self.publication_name,
                    Payload=Ack(AckMessageID=message.Header.MessageId),
                ),
            )

    def start_ping_tasks(self) -> list[asyncio.Task]:
        return [
            asyncio.create_task(
                self.send_ping(link_name), name=f"send_ping<{link_name}>"
            )
            for link_name in self.link_names()
        ]

    async def send_ping(self, link_name: str):
        while not self._states.stopped(link_name):
            message_times = self._message_times.get_copy(link_name)
            link_state = self._states[link_name]
            if (
                message_times.time_to_send_ping(self._settings.mqtt_link_poll_seconds)
                and link_state.active_for_send()
            ):
                self.publish_message(link_name, PingMessage(Src=self.publication_name))
            await asyncio.sleep(
                message_times.seconds_until_next_ping(
                    self._settings.mqtt_link_poll_seconds
                )
            )

    def update_recv_time(self, link_name: str) -> None:
        self._message_times.update_recv(link_name)

    def _recv_activated(self, transition: Transition):
        if transition.link_name == self.upstream_client:
            self._start_reupload()
        self.generate_event(PeerActiveEvent(PeerName=transition.link_name))

    def process_mqtt_suback(
        self, message: Message[MQTTSubackPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        self._logger.path(
            "++LinkManager.process_mqtt_suback client:%s", message.Payload.client_name
        )
        path_dbg = 0
        state_result = self._states.process_mqtt_suback(
            message.Payload.client_name,
            self._mqtt_clients.handle_suback(message.Payload),
        )
        if state_result.is_ok():
            path_dbg |= 0x00000001
            if state_result.value:
                path_dbg |= 0x00000002
                self._logger.comm_event(str(state_result.value))
            if state_result.value.send_activated():
                path_dbg |= 0x00000004
                self.generate_event(
                    MQTTFullySubscribedEvent(PeerName=message.Payload.client_name)
                )
                self.publish_message(
                    message.Payload.client_name,
                    PingMessage(Src=self.publication_name),
                )
            if state_result.value.recv_activated():
                path_dbg |= 0x00000008
                self._recv_activated(state_result.value)
        self._logger.path(
            "--LinkManager.process_mqtt_suback:%d  path:0x%08X",
            state_result.is_ok(),
            path_dbg,
        )
        return state_result



"""Settings for the GridWorks Scada, readable from environment and/or from env files."""

from gwproactor.config.logging import DEFAULT_BYTES_PER_LOG_FILE
from gwproactor.config.logging import DEFAULT_FRACTIONAL_SECOND_FORMAT
from gwproactor.config.logging import DEFAULT_LOG_FILE_NAME
from gwproactor.config.logging import DEFAULT_LOGGING_FORMAT
from gwproactor.config.logging import DEFAULT_NUM_LOG_FILES
from gwproactor.config.logging import FormatterSettings
from gwproactor.config.logging import LoggerLevels
from gwproactor.config.logging import LoggingSettings
from gwproactor.config.logging import RotatingFileHandlerSettings
from gwproactor.config.mqtt import MQTTClient
from gwproactor.config.paths import DEFAULT_BASE_DIR
from gwproactor.config.paths import DEFAULT_BASE_NAME
from gwproactor.config.paths import DEFAULT_LAYOUT_FILE
from gwproactor.config.paths import DEFAULT_NAME
from gwproactor.config.paths import DEFAULT_NAME_DIR
from gwproactor.config.paths import Paths
from gwproactor.config.proactor_settings import ProactorSettings


DEFAULT_ENV_FILE = ".env"

__all__ = [
    # logging
    "DEFAULT_LOGGING_FORMAT",
    "DEFAULT_FRACTIONAL_SECOND_FORMAT",
    "DEFAULT_LOG_FILE_NAME",
    "DEFAULT_BYTES_PER_LOG_FILE",
    "DEFAULT_NUM_LOG_FILES",
    "FormatterSettings",
    "LoggerLevels",
    "LoggingSettings",
    "RotatingFileHandlerSettings",
    # mqtt
    "MQTTClient",
    # paths
    "DEFAULT_ENV_FILE",
    "DEFAULT_BASE_DIR",
    "DEFAULT_BASE_NAME",
    "DEFAULT_NAME",
    "DEFAULT_NAME_DIR",
    "DEFAULT_LAYOUT_FILE",
    "Paths",
    # proactor
    "ProactorSettings",
]



import asyncio
import threading
import time
from typing import Coroutine
from typing import Optional
from typing import Sequence

from gwproto import Message
from result import Result

from gwproactor import ProactorLogger
from gwproactor.actors.actor import MonitoredName
from gwproactor.message import KnownNames
from gwproactor.message import PatInternalWatchdogMessage
from gwproactor.message import ShutdownMessage
from gwproactor.proactor_interface import INVALID_IO_TASK_HANDLE
from gwproactor.proactor_interface import Communicator
from gwproactor.proactor_interface import IOLoopInterface
from gwproactor.proactor_interface import ServicesInterface
from gwproactor.problems import Problems
from gwproactor.sync_thread import SyncAsyncInteractionThread


class IOLoop(Communicator, IOLoopInterface):

    _io_loop: asyncio.AbstractEventLoop
    _io_thread: Optional[threading.Thread] = None
    _tasks: dict[int, Optional[asyncio.Task]]
    _id2task: dict[asyncio.Task, int]
    _completed_tasks: dict[int, asyncio.Task]
    _lock: threading.RLock
    _stop_requested: bool = False
    _next_id = INVALID_IO_TASK_HANDLE + 1
    _pat_timeout: float = SyncAsyncInteractionThread.PAT_TIMEOUT
    _last_pat_time: float = 0.0
    _lg: ProactorLogger

    def __init__(self, services: ServicesInterface) -> None:
        super().__init__(KnownNames.io_loop_manager.value, services)
        self._lg = services.logger
        self._lock = threading.RLock()
        self._tasks = dict()
        self._id2task = dict()
        self._completed_tasks = dict()
        self._io_loop = asyncio.new_event_loop()

    def add_io_coroutine(self, coro: Coroutine, name: str = "") -> int:
        with self._lock:
            if self._stop_requested:
                return INVALID_IO_TASK_HANDLE
            task_id = self._next_id
            self._next_id += 1
            self._tasks[task_id] = None
        self._io_loop.call_soon_threadsafe(self._add_task, coro, name, task_id)
        return task_id

    def cancel_io_routine(self, handle: int) -> None:
        with self._lock:
            task = self._tasks.pop(handle, None)
        if task is not None:
            self._io_loop.call_soon_threadsafe(task.cancel)

    def _add_task(self, coro: Coroutine, name: str, task_id: int) -> None:
        if not name:
            name = coro.__name__
        task = self._io_loop.create_task(
            coro,
            name=name,
        )
        with self._lock:
            self._tasks[task_id] = task
            self._id2task[task] = task_id

    def start(self) -> None:
        self._io_thread = threading.Thread(
            name=self.name, target=self._thread_run, daemon=True
        )
        self._io_thread.start()

    def _started_tasks(self) -> list[asyncio.Task]:
        with self._lock:
            return [task for task in self._tasks.values() if task is not None]

    def stop(self) -> None:
        with self._lock:
            self._stop_requested = True
            tasks = self._started_tasks()
        for task in tasks:
            self._io_loop.call_soon_threadsafe(task.cancel)

    async def join(self):
        pass
        # this often generates errors, although tests pass:
        # await async_polling_thread_join(self._io_thread)

    def _thread_run(self) -> None:
        try:
            self._io_loop.run_until_complete(self._async_run())
        except BaseException as e:  # noqa
            try:
                summary = f"Unexpected IOLoop exception <{e}>"
            except:  # noqa
                summary = "Unexpected IOLoop exception"
            try:
                self._services.send_threadsafe(
                    Message(
                        Payload=Problems(errors=[e]).problem_event(
                            summary=summary, src=self.name
                        )
                    )
                )
            except:  # noqa
                pass
            try:
                self._services.send_threadsafe(
                    ShutdownMessage(
                        Src=self.name,
                        Reason=summary,
                    )
                )
            except:  # noqa
                pass
        finally:
            self._io_loop.stop()

    async def _async_run(self):
        try:
            while not self._stop_requested:
                with self._lock:
                    tasks = self._started_tasks()
                if not self._stop_requested and not tasks:
                    try:
                        await asyncio.sleep(1)
                    except asyncio.CancelledError:
                        pass
                    except GeneratorExit:
                        pass
                else:
                    done, running = await asyncio.wait(
                        tasks, timeout=1.0, return_when="FIRST_COMPLETED"
                    )
                    if self._stop_requested:
                        break
                    with self._lock:
                        for task in done:
                            task_id = self._id2task.pop(task)
                            self._completed_tasks[task_id] = self._tasks.pop(task_id)
                    errors = []
                    for task in done:
                        if not task.cancelled():
                            try:
                                exception = task.exception()
                                if exception is not None:
                                    errors.append(exception)
                            except BaseException as retrieve_exception:
                                errors.append(retrieve_exception)
                    if errors:
                        raise Problems(
                            f"IOLoop caught {len(errors)}.",
                            errors=errors,
                        )
                if self.time_to_pat():
                    self.pat_watchdog()
        finally:
            self._io_loop.stop()

    def time_to_pat(self) -> bool:
        return time.time() >= (self._last_pat_time + (self._pat_timeout / 2))

    def pat_watchdog(self):
        if not self._stop_requested:
            self._last_pat_time = time.time()
            self._services.send_threadsafe(PatInternalWatchdogMessage(src=self.name))

    def process_message(self, message: Message) -> Result[bool, BaseException]:
        raise ValueError("IOLoop does not currently process any messages")

    def _send(self, message: Message) -> None:
        """Ensure this routine from base class does not make unsafe call."""
        self._services.send_threadsafe(message)

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        return [MonitoredName(self.name, self._pat_timeout)]
